@@ -91,13 +91,13 @@ class Mamba2Block(nn.Module):
         # Apply convolution
         if cuda_causal_conv1d is not None:
             xBC = cuda_causal_conv1d(
-                xBC.transpose(1, 2),
+                xBC.transpose(1, 2).contiguous(),
                 rearrange(self.conv1d.weight, "d 1 w -> d w"),
                 bias=self.conv1d.bias,
                 activation="silu",
             ).transpose(1, 2)
         else:
-            xBC = F.silu(self.conv1d(xBC.transpose(1, 2)).transpose(1, 2)[:, :padded_seq_len])
+            xBC = F.silu(self.conv1d(xBC.transpose(1, 2).contiguous()).transpose(1, 2)[:, :padded_seq_len])
         
         x, B, C = torch.split(
             xBC,
@@ -209,6 +209,7 @@ class Mamba2(nn.Module):
                  d_state=None, d_conv=4, expand=2,
                  use_embedding=True, weight_tying=False,
                  bidirectional=False,
+                 pos_encoding=False, pos_encoding_max_len=None,
                  chunk_size=64, device="cpu"):
         super().__init__()
         self.emb_dim = emb_dim
@@ -226,6 +227,12 @@ class Mamba2(nn.Module):
         else:
             self.embedding = nn.Linear(input_dim, emb_dim, bias=False, device=device)
         
+        self.abs_pos_encoding = None
+        if pos_encoding:
+            assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
+            self.pos_encoding_max_len = pos_encoding_max_len
+            self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim, device=device)
+        
         self.layers = nn.ModuleList([
             nn.ModuleDict(
                 dict(
@@ -252,9 +259,12 @@ class Mamba2(nn.Module):
         self.to(device)
 
     def forward(self, x):
-        seq_len = x.shape[1]
+        seq_len = x.size(1)
         if self.use_embedding: x = self.embedding(x.long())
         else: x = self.embedding(x)
+        if self.abs_pos_encoding is not None:
+            pos = torch.arange(seq_len, device=x.device, dtype=torch.long).unsqueeze(0).expand(x.size(0), -1)
+            x = x + self.abs_pos_encoding(pos)
         for layer in self.layers:
             x_norm = layer.norm(x)
             y_f = layer.mixer(x_norm)
@@ -268,7 +278,7 @@ class SeqLinear(nn.Module):
     """
     Linear layer with weights derived from a weighted sum of rank-1 sequential updates.
     """
-    def __init__(self, d_model, n_heads, bias=True,
+    def __init__(self, d_model, n_heads,
                  W_0=False, d_conv=4,
                  u_dim=None, v_dim=None,
                  device="cpu"):
@@ -277,7 +287,6 @@ class SeqLinear(nn.Module):
         self.u_dim = u_dim if u_dim is not None else d_model
         self.v_dim = v_dim if v_dim is not None else d_model
         self.n_heads = n_heads
-        self.bias = bias
         self.device = device
         
         d_in_proj = 2 * self.u_dim + self.v_dim + self.n_heads
@@ -289,10 +298,8 @@ class SeqLinear(nn.Module):
             kernel_size=d_conv,
             groups=d_conv_c,
             padding=d_conv-1,
-            bias=bias,
             device=device,
         )
-        self.dt_rate = nn.Parameter(torch.empty(n_heads, device=device))
         self.dt_bias = nn.Parameter(torch.empty(n_heads, device=device))
         self.W_0 = nn.Parameter(torch.empty(n_heads, self.u_dim // n_heads, self.v_dim // n_heads, device=device)) if W_0 else None
         self.out_proj = nn.Linear(self.v_dim, d_model, bias=False, device=device)
@@ -302,9 +309,8 @@ class SeqLinear(nn.Module):
     def _reset_parameters(self):
         nn.init.xavier_uniform_(self.in_proj.weight)
         nn.init.xavier_uniform_(self.out_proj.weight)
-        nn.init.kaiming_uniform_(self.conv.weight, nonlinearity='relu')
-        nn.init.uniform_(self.dt_bias, -0.5, 0.5)
-        nn.init.uniform_(self.dt_rate, -0.1, 0.1)
+        nn.init.xavier_uniform_(self.conv.weight)
+        nn.init.uniform_(self.dt_bias, -0.1, 0.1)
         
         if self.conv.bias is not None:
             nn.init.constant_(self.conv.bias, 0.)
@@ -312,9 +318,9 @@ class SeqLinear(nn.Module):
             nn.init.constant_(self.in_proj.bias, 0.)
         if self.out_proj.bias is not None:
             nn.init.constant_(self.out_proj.bias, 0.)
-        
-        if self.W_0 is not None: nn.init.constant_(self.W_0, 0.)
-    
+        if self.W_0 is not None:
+            nn.init.constant_(self.W_0, 0.)
+
     def bidirectional(self, x):
         bsz, src_len, d_model = x.size()
         
@@ -334,17 +340,17 @@ class SeqLinear(nn.Module):
                 x.transpose(1, 2).contiguous(),
                 self.conv.weight.squeeze(1),
                 bias=self.conv.bias,
-                activation="silu",
+                activation=None,
             ).transpose(1, 2)
             x_b = cuda_causal_conv1d(
                 x.flip(1).transpose(1, 2).contiguous(),
                 self.conv.weight.squeeze(1),
                 bias=self.conv.bias,
-                activation="silu",
+                activation=None,
             ).transpose(1, 2)
         else:
-            x_f = F.silu(self.conv(x.transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len])
-            x_b = F.silu(self.conv(x.flip(1).transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len])
+            x_f = self.conv(x.transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len]
+            x_b = self.conv(x.flip(1).transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len]
         
         z_f, u_f, v_f = torch.split(
             x_f,
@@ -364,19 +370,22 @@ class SeqLinear(nn.Module):
             ],
             dim=-1
         )
-        
-        t = torch.exp(torch.linspace(0, 1, src_len, device=self.device).unsqueeze(-1) * self.dt_rate)
-        dt_f = F.softplus(dt + self.dt_bias) * t
+                
+        log_dt_f = torch.cumsum(-F.leaky_relu(dt + self.dt_bias), dim=1)
+        log_dt_f = log_dt_f - log_dt_f.topk(1, dim=1).values
+        dt_f = torch.exp(log_dt_f)
         z_f = rearrange(z_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         u_f = rearrange(u_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         v_f = rearrange(v_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         
-        dt_b = F.softplus(dt + self.dt_bias).flip(1) * t
+        log_dt_b = torch.cumsum(-F.leaky_relu(dt.flip() + self.dt_bias), dim=1)
+        log_dt_b = log_dt_b - log_dt_b.topk(1, dim=1).values
+        dt_b = torch.exp(log_dt_b)
         z_b = rearrange(z_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         u_b = rearrange(u_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         v_b = rearrange(v_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         
-        W_f = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt, u_f, v_f), dim=1)
+        W_f = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt_f, u_f, v_f), dim=1)
         W_b = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt_b, u_b, v_b), dim=1)
         
         if self.W_0 is not None:
@@ -388,6 +397,7 @@ class SeqLinear(nn.Module):
         
         out_f = torch.matmul(z_f.unsqueeze(-2), W_f).squeeze(-2) / norm_f
         out_b = torch.matmul(z_b.unsqueeze(-2), W_b).squeeze(-2) / norm_b
+        
         out = rearrange(out_f + out_b, 'b s h d -> b s (h d)')
         out = self.out_proj(out)
         return out
@@ -412,10 +422,10 @@ class SeqLinear(nn.Module):
                 x.transpose(1, 2).contiguous(),
                 self.conv.weight.squeeze(1),
                 bias=self.conv.bias,
-                activation="silu",
+                activation=None,
             ).transpose(1, 2)
         else:
-            x = F.silu(self.conv(x.transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len])
+            x = self.conv(x.transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len]
         
         z, u, v = torch.split(
             x,
@@ -427,22 +437,23 @@ class SeqLinear(nn.Module):
             dim=-1
         )
         
-        t = torch.exp(torch.linspace(0, 1, src_len, device=self.device).unsqueeze(-1) * self.dt_rate)
-        dt = F.softplus(dt + self.dt_bias) * t
+        log_dt = torch.cumsum(-F.leaky_relu(dt + self.dt_bias), dim=1)
+        log_dt = log_dt - log_dt.topk(1, dim=1).values
+        dt = torch.exp(log_dt)
         z = rearrange(z, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         u = rearrange(u, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         v = rearrange(v, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         
         if mode == "causal":
+            norm = torch.cumsum(dt, dim=1).unsqueeze(-1)
             W = torch.cumsum(torch.einsum('bsh, bshd, bshe -> bshde', dt, u, v), dim=1)
         elif mode == "noncausal":
+            norm = torch.sum(dt, dim=1, keepdim=True).unsqueeze(-1)
             W = torch.einsum('bsh, bshd, bshe -> bhde', dt, u, v).unsqueeze(1)
         
         if self.W_0 is not None: W = W + self.W_0.unsqueeze(1)
         
-        norm = torch.cumsum(dt, dim=1).unsqueeze(-1)
-        
-        out = torch.matmul(z.unsqueeze(-2), W).squeeze(-2) / norm
+        out = torch.einsum('bshd, bshde -> bshe', z, W) / norm
         out = rearrange(out, 'b s h d -> b s (h d)')
         out = self.out_proj(out)
         return out
@@ -453,7 +464,7 @@ class SeqMLP(nn.Module):
                  u_dim=None, v_dim=None, W_0=False,
                  d_conv=4, pos_encoding=False, pos_encoding_max_len=None,
                  use_embedding=True, weight_tying=False,
-                 mode="causal", bias=True, device="cpu"):
+                 mode="causal", device="cpu"):
         super().__init__()
         self.emb_dim = emb_dim
         self.input_dim = input_dim
@@ -485,7 +496,6 @@ class SeqMLP(nn.Module):
                         v_dim=v_dim,
                         W_0=W_0,
                         d_conv=d_conv,
-                        bias=bias,
                         device=device
                     ),
                     dropout = nn.Dropout(dropout),
